#include <list>
#include <string>
#include <algorithm>

#include <regex.h>

#include <arc/common.h>
#include <arc/joblist.h>
#include <arc/notify.h>

std::list<std::string> GetJobIDs(bool all,
                                const std::list<std::string>& jobs,
                                const std::list<std::string>& clusterselect,
                                const std::list<std::string>& clusterreject) {

	bool allbut = false;
	if (jobs.empty() && clusterselect.empty() &&
		(!clusterreject.empty())) allbut = true;

	std::list<std::string> jobids;

	std::string filename = GetEnv("HOME");
	filename.append("/.ngjobs");

	LockFile(filename);

	std::ifstream ngjobs(filename.c_str());

	if (all ||
		(jobs.empty() && clusterselect.empty() && !clusterreject.empty()) ||
		!clusterselect.empty()) {
		std::string jobidname;
		while (getline(ngjobs, jobidname)) {
			size_t pos = jobidname.find('#');
			std::string jobid = jobidname.substr(0, pos);
			std::string jobname = jobidname.substr(pos + 1);
			if (!all && !clusterselect.empty()) {
				size_t pos0 = (jobid.substr(0, 9) == "gsiftp://" ? 9 : 0);
				size_t pos = jobid.find_first_of(":/", pos0);
				std::string clustername;
				if (pos == std::string::npos)
					clustername = jobid.substr(pos0);
				else
					clustername = jobid.substr(pos0, pos - pos0);
				if (find(clusterselect.begin(), clusterselect.end(),
				         clustername) == clusterselect.end())
					continue;
			}
			if (find(jobids.begin(), jobids.end(), jobid) == jobids.end())
				jobids.push_back(jobid);
		}
	}

	for (std::list<std::string>::const_iterator vsci = jobs.begin();
	     vsci != jobs.end(); vsci++) {
		bool found = false;
		ngjobs.clear();
		ngjobs.seekg(std::ios::beg);
		std::string jobidname;
		while (getline(ngjobs, jobidname)) {
			int pos = jobidname.find('#');
			std::string jobid = jobidname.substr(0, pos);
			std::string jobname = jobidname.substr(pos + 1);
			bool regmatch = false;
			if (vsci->find_first_of("*?") != std::string::npos) {
				std::string reg = '^' + *vsci + '$';
				size_t pos = 0;
				while ((pos = reg.find('.', pos)) != std::string::npos) {
					reg.insert(pos, "\\");
					pos += 2;
				}
				pos = 0;
				while ((pos = reg.find('*', pos)) != std::string::npos) {
					reg.insert(pos, ".");
					pos += 2;
				}
				pos = 0;
				while ((pos = reg.find('?', pos)) != std::string::npos) {
					reg.replace(pos, 1, ".");
					pos++;
				}
				regex_t regx;
				regcomp(&regx, reg.c_str(), 0);
				if (regexec(&regx, jobid.c_str(), 0, NULL, 0) == 0 ||
					regexec(&regx, jobname.c_str(), 0, NULL, 0) == 0)
					regmatch = true;
				regfree(&regx);
			}
			if (regmatch || jobid == *vsci || jobname == *vsci) {
				if (find(jobids.begin(), jobids.end(), jobid) == jobids.end())
					jobids.push_back(jobid);
				found = true;
			}
		}
		if (!found) {
			if (find(jobids.begin(), jobids.end(), *vsci) == jobids.end())
				jobids.push_back(*vsci);
		}
	}

	ngjobs.close();

	UnlockFile(filename);

	if (!clusterreject.empty()) {
		std::list<std::string> filteredjobids;
		for (std::list<std::string>::iterator vsi = jobids.begin();
		     vsi != jobids.end(); vsi++) {
			size_t pos0 = 0;
			if (vsi->substr(0, 9) == "gsiftp://") pos0 = 9;
			size_t pos = vsi->find_first_of(":/", pos0);
			std::string clustername;
			if (pos == std::string::npos)
				clustername = vsi->substr(pos0);
			else
				clustername = vsi->substr(pos0, pos - pos0);
			if (find(clusterreject.begin(), clusterreject.end(),
			         clustername) == clusterreject.end())
				filteredjobids.push_back(*vsi);
		}
		jobids = filteredjobids;
	}

	return jobids;
}


void AddJobID(const std::string& jobid, const std::string& jobname) {

	std::string filename = GetEnv("HOME");
	filename.append("/.ngjobs");

	LockFile(filename);

	std::ifstream oldngjobs(filename.c_str());
	std::string newfilename = filename + ".tmp";
	std::ofstream newngjobs(newfilename.c_str());
	std::string jobidname;
	bool written = false;
	while (getline (oldngjobs, jobidname)) {
		int pos = jobidname.find('#');
		std::string jobname_ = jobidname.substr (pos + 1);
		if (!written && jobname_ > jobname) {
			newngjobs << jobid << '#' << jobname << std::endl;
			written = true;
		}
		newngjobs << jobidname << std::endl;
	}
	if (!written) newngjobs << jobid << '#' << jobname << std::endl;
	oldngjobs.close();
	newngjobs.close();
	remove (filename.c_str());
	rename (newfilename.c_str(), filename.c_str());

	UnlockFile(filename);
}


void RemoveJobID(const std::string& jobid) {

	std::string filename = GetEnv("HOME");
	filename.append("/.ngjobs");

	LockFile(filename);

	std::ifstream oldngjobs(filename.c_str());
	std::string newfilename = filename + ".tmp";
	std::ofstream newngjobs(newfilename.c_str());
	std::string jobidname;
	while (getline(oldngjobs, jobidname)) {
		int pos = jobidname.find('#');
		std::string jobid_ = jobidname.substr(0, pos);
		if (jobid_ != jobid) newngjobs << jobidname << std::endl;
	}
	oldngjobs.close();
	newngjobs.close();
	remove(filename.c_str());
	rename(newfilename.c_str(), filename.c_str());

	UnlockFile(filename);
}
