#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <arc/certificate.h>
#include <arc/notify.h>
#include <arc/ldapquery.h>
#include <arc/mdsquery.h>
#include <arc/stringconv.h>
#include <arc/xrsl.h>
#include <arc/common.h>
#include <arc/jobsubmission.h>

#include <string>
#include <iostream>
#include <list>
#include <algorithm>

using namespace std;

void FindVOMembers(const string& attr, const string& value, void* arg);

int main(int argc, char* argv[]) {
	string progname = argv[0];
	string::size_type pos = progname.rfind('/');
	if (pos!=string::npos)
		progname = progname.substr(pos+1);

	if (argc<2) {
		cout << "Error: Not enough arguments given." << endl;
		return 1;
	}

	if (argc>3) {
		cout << "Error: Too many arguments given." << endl;
		return 1;
	}

	SetNotifyLevel(FATAL);

	string arg1(argv[1]);
	if (arg1=="certificate") {
		// Run certificate test
		cout << "Certificate information:" << endl;
		cout << "========================" << endl;
		cout << endl;

		string issuersn;
		try {
			Certificate user;
			cout << "Certificate  : " << user.GetCertFilename() << endl;
			cout << "Subject name : " << user.GetSN() << endl;

			Time::SetFormat(ASCTime);
			cout << "Valid until  : ";
			if (user.IsExpired()) {
				cout << "Expired" << endl;
			} else {
				cout << user.ExpiryTime() << endl;
			}

			issuersn = user.GetIssuerSN();
		} catch(CertificateError e) {
			cout << "No user-certificate found." << endl;
		}

		cout << endl;

		try {
			Certificate proxy(PROXY);
			cout << "Proxy        : " << proxy.GetCertFilename() << endl;
			cout << "Proxy-subject: " << proxy.GetSN() << endl;

			cout << "Valid for    : ";
			if (proxy.IsExpired()) {
				cout << "Expired" << endl;
			} else {
				cout << proxy.ValidFor() << endl;
			}
		} catch(CertificateError e) {
			cout << "No proxy found." << endl;
		}
	
		cout << endl;

		if (!issuersn.empty()) {
			cout << "Certificate issuer: " << issuersn << endl;
			cout << endl;
		}

		cout << "CA-certificates installed:" << endl;
		list<Certificate> calist = GetCAList();
		if (calist.size()==0)
			cout << "None." << endl;

		list<Certificate>::iterator it;
		for (it = calist.begin(); it != calist.end(); it++)
			cout << it->GetSN() << endl;

		if (!issuersn.empty() && !CheckIssuer(issuersn)) {
			cout << endl;
			cout << "Warning: Your issuer's certificate is not installed." << endl;
		}

		return 0;
	}

	if (arg1=="vo") {

		cout << "VO-information:" << endl;
		cout << "===============" << endl;
		cout << endl;

		string usersn;
		try {
			Certificate user;
			usersn = user.GetSN();
		} catch(CertificateError e) {
			cout << "Error: No user-certificate found." << endl;
			return 1;
		}

		cout << "Certificate-subject: " << usersn << endl << endl;
		
		list<pair<string, URL> > vos;

		URL NGVO("ldap://grid-vo.nordugrid.org/dc=org/dc=nordugrid/ou=people");
		string NGVOdesc("NorduGrid VO      ");

		URL NGguestVO("ldap://grid-vo.nordugrid.org/dc=org/dc=nordugrid/ou=guest");
		string NGguestVOdesc("NorduGrid guest VO");

		//URL ARCVO("ldap://grid-vo.nordugrid.org/dc=org/dc=nordugrid/ou=people");
		//string ARCVOdesc("ARC Community VO ");

		vos.push_back(make_pair(NGVOdesc, NGVO));
		vos.push_back(make_pair(NGguestVOdesc, NGguestVO));
		//vos.push_back(make_pair(ARCVOdesc, ARCVO));

		list<pair<string, URL> >::iterator it;
		for (it = vos.begin(); it != vos.end(); it++) {
			cout << it->first << ": ";
			LdapQuery ldapq(it->second.Host(), it->second.Port());
			list<string> vomembers;
			try {
				ldapq.Query(it->second.BaseDN());
				ldapq.Result(FindVOMembers, (void*)&vomembers);

				if (find(vomembers.begin(), vomembers.end(), usersn)==vomembers.end()) {
					cout << "Not a member." << endl;
				} else {
					cout << "Member." << endl;
				}
			} catch (LdapQueryError e) {
				cout << "Can't contact LDAP-server." << endl;
			}
		}

		return 0;
	}

	if (arg1=="resources") {

		cout << "Resource authorizations:" << endl;
		cout << "========================" << endl;
		cout << endl;

		string usersn;
		try {
			Certificate user;
			usersn = user.GetSN();
		} catch(CertificateError e) {
			cout << "Error: No user-certificate found." << endl;
			return 1;
		}

		cout << "Certificate-subject: " << usersn << endl << endl;
		cout << "Retrieving information ..." << endl << endl;

		// clusters
		list<string> authclusters;
		// The following line is not good on RH73 - bad constructor?
		// list<Queue> queueinfo = GetQueueInfo();
		list<Queue> queueinfo;
		queueinfo = GetQueueInfo();
		list<Queue>::iterator it;
		for (it = queueinfo.begin(); it != queueinfo.end(); it++) {
			if (it->users.size()>0)
				authclusters.push_back(it->cluster.hostname);
		}

		authclusters.sort();
		authclusters.unique();

		cout << "You are authorized at the following clusters:" << endl;
		if (authclusters.empty())
			cout << "None.";

		list<string>::iterator strit;
		for (strit = authclusters.begin(); strit != authclusters.end(); strit++)
			cout << *strit << "  ";

		cout << endl;

		// storage-elements
		cout << endl;

		list<StorageElement> seinfo = GetSEInfo();
		list<StorageElement>::iterator seit;
		cout << "You are authorized at the following storage-elements:" << endl;
		if (seinfo.empty())
			cout << "None." << endl;

		for (seit = seinfo.begin(); seit != seinfo.end(); seit++) {
			list<string> authusers = seit->auth_users;
			if (find(authusers.begin(), authusers.end(), usersn)!=authusers.end())
				cout << seit->name << "  ";
		}

		cout << endl;
		return 0;

	}

	if (arg1=="job") {
		unsigned int arg2 = 1;
		if (argc<3) {
			cout << "No test-job-id given. Submitting default test-job" << endl;
		} else {
			try {
				arg2 = stringtoui(argv[2]);
			} catch(StringConvError e) {
				cout << "Illegal test-job-id: " << e.what() << endl;
				return 1;
			}
		}

		Xrsl xrsl;
		// default test-job
		if (arg2==1) xrsl = Xrsl("&(executable=run.sh)(arguments=5)(inputfiles=(\"run.sh\" \"http://www.nordugrid.org;cache=no/data/run.sh\")(\"Makefile\" \"rls://gridsrv3.nbi.dk/Makefile\")(\"prime.cpp\" \"ftp://ftp.nordugrid.org;cache=no/applications/test/prime.cpp\"))(stderr=primenumbers)(outputfiles=(\"primenumbers\" \"\"))(jobname=ngtest-job-1)(stdout=stdout)(gmlog=gmlog)(CPUTime=5)");
		if (arg2==2) xrsl = Xrsl("&(executable=/bin/echo)(arguments=\"hello, grid\")(jobname=ngtest-job-2)(cputime=5)");
		if (arg2==3) xrsl = Xrsl("&(executable=md5sum.sh)(arguments=downloaded[0-9])(inputfiles=(\"downloaded0\" \"http://www.nordugrid.org;cache=yes/data/echo.sh\")(\"downloaded1\" \"http://www.nordugrid.org;cache=yes/data/somefile\")(\"downloaded2\" \"gsiftp://hathi.hep.lu.se;cache=yes/public/ngtest/echo.sh\")(\"downloaded3\" \"gsiftp://hathi.hep.lu.se;cache=yes/public/ngtest/somefile\")(\"downloaded4\" \"ftp://ftp.nordugrid.org;cache=yes/applications/test/echo.sh\")(\"downloaded5\" \"ftp://ftp.nordugrid.org;cache=yes/applications/test/somefile\")(\"downloaded6\" \"rls://gridsrv3.nbi.dk/echo.sh\")(\"downloaded7\" \"rls://gridsrv3.nbi.dk/somefile\")(\"md5sum.sh\" \"ftp://ftp.nordugrid.org;cache=yes/applications/test/md5sum.sh\"))(jobName=ngtest-job-3)(stdout=stdout)(stderr=stderr)(gmlog=gmlog)(CPUTime=5)");

		cout << "Submitting test-job " << arg2 << ": " << endl << xrsl.str() << endl;
		
		// Check for valid proxy
		try {
			Certificate proxy(PROXY);
			if (proxy.IsExpired()) {
				cout << "Your proxy is expired. Please create a new one." << endl;
				return 1;
			}
		} catch(CertificateError e) {
			cout << "No proxy found. Please create one." << endl;
			return 1;
		}

		cout << "Client middleware: nordugrid-" << VERSION << ", globus-";
		// Find globus middleware
		string globusloc = GetEnv("GLOBUS_LOCATION");
		if (globusloc.empty()) globusloc = DEFAULT_GLOBUS_LOCATION;
		string versionfile = globusloc + "/share/doc/VERSION";
		ifstream file(versionfile.c_str());
		
		if (file) {
			string version;
			file >> version;
			if (version.empty()) {
				cout << "unknown" << endl;
			} else {
				cout << version << endl;
			}
			file.close();
		}

		// Submit job
		list<Target> targets = PrepareJobSubmission(xrsl);
		string jobid = SubmitJob(xrsl, targets);
		cout << "Job submitted with jobid: " << jobid << endl;
		
		// Get middleware of server
		URL joburl(jobid);
		URL ldapurl("ldap://" + joburl.Host() + ":2135/o=grid/mds-vo-name=local");
		Cluster cluster = GetClusterInfo(ldapurl);
		cout << "Cluster (" << cluster.hostname << ") middleware: ";

		string nordugridmw("nordugrid-unknown");
		string globusmw("globus-unknown");
		list<RuntimeEnvironment> mws = cluster.middlewares;
		list<RuntimeEnvironment>::iterator it;
		for (it = mws.begin(); it != mws.end(); it++) {
			if (it->Name()=="nordugrid") nordugridmw = it->str();
			if (it->Name()=="globus") globusmw = it->str();
		}

		cout << nordugridmw << ", " << globusmw << endl;

		return 0;
	}

	// benchmarking
	// reverse hostname lookup
	// host certificate setup

	cout << "Error: Illegal first argument given" << endl;
	return 1;
}


void FindVOMembers(const string& attr, const string& value, void* arg) {
	if (attr!="description") return;
	list<string>* vomembers = (list<string>*)arg;

	string val(value.substr(8));
	while (val[0]==' ') val = val.substr(1);
	vomembers->push_back(val);
}
